@@ -28,12 +28,8 @@
 #include <php.h>
 #include <php_ini.h>
 #include <ext/standard/info.h>
-#if ZEND_MODULE_API_NO >= 20141001
 #include <ext/standard/php_smart_string.h>
-#else
-#include <ext/standard/php_smart_str.h>
-#endif
-#if PHP_MAJOR_VERSION >= 7 && defined(HAVE_APCU_SUPPORT)
+#if defined(HAVE_APCU_SUPPORT)
 #include <ext/standard/php_var.h>
 #include <ext/apcu/apc_serializer.h>
 #include <zend_smart_str.h>
@@ -53,12 +49,19 @@
 #define ZSTD_CLEVEL_DEFAULT 3
 #endif
 
-#define FRAME_HEADER_SIZE 5
-#define BLOCK_HEADER_SIZE 3
-#define MAX_HEADER_SIZE FRAME_HEADER_SIZE+3
-
 #define DEFAULT_COMPRESS_LEVEL 3
 
+// zend_string_efree doesnt exist in PHP7.2, 20180731 is PHP 7.3
+#if ZEND_MODULE_API_NO < 20180731
+#define zend_string_efree(string) zend_string_free(string)
+#endif
+
+#define ZSTD_WARNING(...) \
+    php_error_docref(NULL, E_WARNING, __VA_ARGS__)
+
+#define ZSTD_IS_ERROR(result) \
+    UNEXPECTED(ZSTD_isError(result))
+
 ZEND_BEGIN_ARG_INFO_EX(arginfo_zstd_compress, 0, 0, 1)
     ZEND_ARG_INFO(0, data)
     ZEND_ARG_INFO(0, level)
@@ -79,99 +82,139 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_zstd_uncompress_dict, 0, 0, 2)
     ZEND_ARG_INFO(0, dictBuffer)
 ZEND_END_ARG_INFO()
 
-ZEND_FUNCTION(zstd_compress)
+static size_t zstd_check_compress_level(long level)
 {
-    zval *data;
-    char *output;
-    size_t size, result;
-    long level = DEFAULT_COMPRESS_LEVEL;
-    uint16_t maxLevel = (uint16_t)ZSTD_maxCLevel();
-
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
-                              "z|l", &data, &level) == FAILURE) {
-        RETURN_FALSE;
-    }
-
-    if (Z_TYPE_P(data) != IS_STRING) {
-        zend_error(E_WARNING, "zstd_compress: expects parameter to be string.");
-        RETURN_FALSE;
-    }
-
+    uint16_t maxLevel = (uint16_t) ZSTD_maxCLevel();
 
 #if ZSTD_VERSION_NUMBER >= 10304
     if (level > maxLevel) {
-      zend_error(E_WARNING, "zstd_compress: compression level (%ld)"
-                 " must be within 1..%d or smaller then 0", level, maxLevel);
+        ZSTD_WARNING("compression level (%ld)"
+            " must be within 1..%d or smaller then 0", level, maxLevel);
+      return 0;
+    }
 #else
     if (level > maxLevel || level < 0) {
-      zend_error(E_WARNING, "zstd_compress: compression level (%ld)"
+      ZSTD_WARNING("compression level (%ld)"
                  " must be within 1..%d", level, maxLevel);
+      return 0;
+    }
 #endif
+    return 1;
+}
+
+// Truncate string to given size
+static zend_always_inline zend_string* zstd_string_output_truncate(zend_string* output, size_t real_length)
+{
+    size_t capacity = ZSTR_LEN(output);
+    size_t free_space = capacity - real_length;
+
+    // Reallocate just when capacity and real size differs a lot or the free space is bigger than 1 MB
+    if (UNEXPECTED(free_space > (capacity / 8) || free_space > (1024 * 1024))) {
+        output = zend_string_truncate(output, real_length, 0);
+    }
+    ZSTR_LEN(output) = real_length;
+    ZSTR_VAL(output)[real_length] = '\0';
+    return output;
+}
+
+ZEND_FUNCTION(zstd_compress)
+{
+    zend_string *output;
+    size_t size, result;
+    long level = DEFAULT_COMPRESS_LEVEL;
+
+    char *input;
+    size_t input_len;
+
+#if PHP_VERSION_ID < 80000
+    zval *data;
+    if (zend_parse_parameters(ZEND_NUM_ARGS(),
+                              "z|l", &data, &level) == FAILURE) {
       RETURN_FALSE;
     }
+    if (Z_TYPE_P(data) != IS_STRING) {
+      zend_error(E_WARNING, "zstd_compress(): expects parameter to be string.");
+      RETURN_FALSE;
+    }
+    input = Z_STRVAL_P(data);
+    input_len = Z_STRLEN_P(data);
+#else
+    ZEND_PARSE_PARAMETERS_START(1, 2)
+        Z_PARAM_STRING(input, input_len)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(level)
+    ZEND_PARSE_PARAMETERS_END();
+#endif
 
-    size = ZSTD_compressBound(Z_STRLEN_P(data));
-    output = (char *)emalloc(size + 1);
-    if (!output) {
-        zend_error(E_WARNING, "zstd_compress: memory error");
+    if (!zstd_check_compress_level(level)) {
         RETURN_FALSE;
     }
 
-    result = ZSTD_compress(output, size, Z_STRVAL_P(data), Z_STRLEN_P(data),
+    size = ZSTD_compressBound(input_len);
+    output = zend_string_alloc(size, 0);
+
+    result = ZSTD_compress(ZSTR_VAL(output), size, input, input_len,
                            level);
 
-    if (ZSTD_isError(result)) {
+    if (ZSTD_IS_ERROR(result)) {
+        zend_string_efree(output);
         RETVAL_FALSE;
-    } else if (result <= 0) {
-        RETVAL_FALSE;
-    } else {
-#if ZEND_MODULE_API_NO >= 20141001
-        RETVAL_STRINGL(output, result);
-#else
-        RETVAL_STRINGL(output, result, 1);
-#endif
     }
 
-    efree(output);
+    output = zstd_string_output_truncate(output, result);
+    RETVAL_NEW_STR(output);
 }
 
 ZEND_FUNCTION(zstd_uncompress)
 {
-    zval *data;
     uint64_t size;
     size_t result;
-    void *output;
+    zend_string *output;
     uint8_t streaming = 0;
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
+    char *input;
+    size_t input_len;
+
+#if PHP_VERSION_ID < 80000
+    zval *data;
+    if (zend_parse_parameters(ZEND_NUM_ARGS(),
                               "z", &data) == FAILURE) {
-        RETURN_FALSE;
+      RETURN_FALSE;
     }
-
     if (Z_TYPE_P(data) != IS_STRING) {
-        zend_error(E_WARNING,
-                   "zstd_uncompress: expects parameter to be string.");
-        RETURN_FALSE;
+      zend_error(E_WARNING,
+                 "zstd_uncompress(): expects parameter to be string.");
+      RETURN_FALSE;
     }
+    input = Z_STRVAL_P(data);
+    input_len = Z_STRLEN_P(data);
+#else
+    ZEND_PARSE_PARAMETERS_START(1, 1)
+        Z_PARAM_STRING(input, input_len)
+    ZEND_PARSE_PARAMETERS_END();
+#endif
 
-    size = ZSTD_getFrameContentSize(Z_STRVAL_P(data), Z_STRLEN_P(data));
+    size = ZSTD_getFrameContentSize(input, input_len);
     if (size == ZSTD_CONTENTSIZE_ERROR) {
-        zend_error(E_WARNING, "zstd_uncompress: it was not compressed by zstd");
+        ZSTD_WARNING("it was not compressed by zstd");
         RETURN_FALSE;
     } else if (size == ZSTD_CONTENTSIZE_UNKNOWN) {
         streaming = 1;
         size = ZSTD_DStreamOutSize();
     }
 
-    output = emalloc(size);
-    if (!output) {
-        zend_error(E_WARNING, "zstd_uncompress: memory error");
-        RETURN_FALSE;
-    }
+    output = zend_string_alloc(size, 0);
 
     if (!streaming) {
-        result = ZSTD_decompress(output, size,
-                                 Z_STRVAL_P(data), Z_STRLEN_P(data));
+        result = ZSTD_decompress(ZSTR_VAL(output), size,
+                                 input, input_len);
+
+        if (ZSTD_IS_ERROR(result)) {
+            zend_string_efree(output);
+            ZSTD_WARNING("can not decompress stream");
+            RETURN_FALSE;
+        }
+
     } else {
         ZSTD_DStream *stream;
         ZSTD_inBuffer in = { NULL, 0, 0 };
@@ -179,40 +222,39 @@ ZEND_FUNCTION(zstd_uncompress)
 
         stream = ZSTD_createDStream();
         if (stream == NULL) {
-            efree(output);
-            zend_error(E_WARNING, "zstd_uncompress: can not create stream");
+            zend_string_efree(output);
+            ZSTD_WARNING("can not create stream");
             RETURN_FALSE;
         }
 
         result = ZSTD_initDStream(stream);
-        if (ZSTD_isError(result)) {
-            efree(output);
+        if (ZSTD_IS_ERROR(result)) {
+            zend_string_efree(output);
             ZSTD_freeDStream(stream);
-            zend_error(E_WARNING, "zstd_uncompress: can not init stream");
+            ZSTD_WARNING("can not init stream");
             RETURN_FALSE;
         }
 
-        in.src = Z_STRVAL_P(data);
-        in.size = Z_STRLEN_P(data);
+        in.src = input;
+        in.size = input_len;
         in.pos = 0;
 
-        out.dst = output;
+        out.dst = ZSTR_VAL(output);
         out.size = size;
         out.pos = 0;
 
         while (in.pos < in.size) {
             if (out.pos == out.size) {
                 out.size += size;
-                output = erealloc(output, out.size);
-                out.dst = output;
+                output = zend_string_extend(output, out.size, 0);
+                out.dst = ZSTR_VAL(output);
             }
 
             result = ZSTD_decompressStream(stream, &out, &in);
-            if (ZSTD_isError(result)) {
-                efree(output);
+            if (ZSTD_IS_ERROR(result)) {
+                zend_string_efree(output);
                 ZSTD_freeDStream(stream);
-                zend_error(E_WARNING,
-                           "zstd_uncompress: can not decompress stream");
+                ZSTD_WARNING("can not decompress stream");
                 RETURN_FALSE;
             }
 
@@ -226,161 +268,114 @@ ZEND_FUNCTION(zstd_uncompress)
         ZSTD_freeDStream(stream);
     }
 
-    if (ZSTD_isError(result)) {
-        RETVAL_FALSE;
-    } else if (result < 0) {
-        RETVAL_FALSE;
-    } else {
-#if ZEND_MODULE_API_NO >= 20141001
-        RETVAL_STRINGL(output, result);
-#else
-        RETVAL_STRINGL(output, result, 1);
-#endif
-    }
-
-    efree(output);
+    output = zstd_string_output_truncate(output, result);
+    RETVAL_NEW_STR(output);
 }
 
 ZEND_FUNCTION(zstd_compress_dict)
 {
-    zval *data, *dictBuffer;
     long level = DEFAULT_COMPRESS_LEVEL;
-    uint16_t maxLevel = (uint16_t) ZSTD_maxCLevel();
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
-                              "zz|l", &data, &dictBuffer, &level) == FAILURE) {
-        RETURN_FALSE;
-    }
-    if (Z_TYPE_P(data) != IS_STRING) {
-        zend_error(E_WARNING, "zstd_compress_dict:"
-                   " expects the first parameter to be string.");
-        RETURN_FALSE;
-    }
-    if (Z_TYPE_P(dictBuffer) != IS_STRING) {
-        zend_error(E_WARNING, "zstd_compress_dict:"
-                   " expects the second parameter to be string.");
+    zend_string *output;
+    char *input, *dict;
+    size_t input_len, dict_len;
+
+    ZEND_PARSE_PARAMETERS_START(2, 3)
+        Z_PARAM_STRING(input, input_len)
+        Z_PARAM_STRING(dict, dict_len)
+        Z_PARAM_OPTIONAL
+        Z_PARAM_LONG(level)
+    ZEND_PARSE_PARAMETERS_END();
+
+    if (!zstd_check_compress_level(level)) {
         RETURN_FALSE;
     }
 
-#if ZSTD_VERSION_NUMBER >= 10304
-    if (level > maxLevel) {
-      zend_error(E_WARNING, "zstd_compress_dict: compression level (%ld)"
-                 " must be within 1..%d or smaller then 0", level, maxLevel);
-#else
-    if (level > maxLevel || level < 0) {
-      zend_error(E_WARNING, "zstd_compress_dict: compression level (%ld)"
-                 " must be within 1..%d", level, maxLevel);
-#endif
-      RETURN_FALSE;
-    }
-
-    size_t const cBuffSize = ZSTD_compressBound(Z_STRLEN_P(data));
-    void* const cBuff = emalloc(cBuffSize);
-    if (!cBuff) {
-        zend_error(E_WARNING, "zstd_compress_dict: memory error");
-        RETURN_FALSE;
-    }
     ZSTD_CCtx* const cctx = ZSTD_createCCtx();
     if (cctx == NULL) {
-        efree(cBuff);
-        zend_error(E_WARNING, "ZSTD_createCCtx() error");
+        ZSTD_WARNING("ZSTD_createCCtx() error");
         RETURN_FALSE;
     }
-    ZSTD_CDict* const cdict = ZSTD_createCDict(Z_STRVAL_P(dictBuffer),
-                                               Z_STRLEN_P(dictBuffer),
+    ZSTD_CDict* const cdict = ZSTD_createCDict(dict,
+                                               dict_len,
                                                level);
     if (!cdict) {
-        efree(cBuff);
-        zend_error(E_WARNING, "ZSTD_createCDict() error");
+        ZSTD_freeCStream(cctx);
+        ZSTD_WARNING("ZSTD_createCDict() error");
         RETURN_FALSE;
     }
-    size_t const cSize = ZSTD_compress_usingCDict(cctx, cBuff, cBuffSize,
-                                                  Z_STRVAL_P(data),
-                                                  Z_STRLEN_P(data),
+
+    size_t const cBuffSize = ZSTD_compressBound(input_len);
+    output = zend_string_alloc(cBuffSize, 0);
+
+    size_t const cSize = ZSTD_compress_usingCDict(cctx, ZSTR_VAL(output), cBuffSize,
+                                                  input,
+                                                  input_len,
                                                   cdict);
-    if (ZSTD_isError(cSize)) {
-        efree(cBuff);
-        zend_error(E_WARNING, "zstd_compress_dict: %s",
-                   ZSTD_getErrorName(cSize));
+    if (ZSTD_IS_ERROR(cSize)) {
+        ZSTD_freeCStream(cctx);
+        ZSTD_freeCDict(cdict);
+        zend_string_efree(output);
+        ZSTD_WARNING("%s", ZSTD_getErrorName(cSize));
         RETURN_FALSE;
     }
+
+    output = zstd_string_output_truncate(output, cSize);
+    RETVAL_NEW_STR(output);
+
     ZSTD_freeCCtx(cctx);
     ZSTD_freeCDict(cdict);
-
-#if ZEND_MODULE_API_NO >= 20141001
-    RETVAL_STRINGL(cBuff, cSize);
-#else
-    RETVAL_STRINGL(cBuff, cSize, 1);
-#endif
-
-    efree(cBuff);
 }
 
 ZEND_FUNCTION(zstd_uncompress_dict)
 {
-    zval *data, *dictBuffer;
+    char *input, *dict;
+    size_t input_len, dict_len;
+    zend_string *output;
 
-    if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC,
-                              "zz", &data, &dictBuffer) == FAILURE) {
-        RETURN_FALSE;
-    }
-    if (Z_TYPE_P(data) != IS_STRING) {
-        zend_error(E_WARNING, "zstd_uncompress_dict:"
-                   " expects the first parameter to be string.");
-        RETURN_FALSE;
-    }
-    if (Z_TYPE_P(dictBuffer) != IS_STRING) {
-        zend_error(E_WARNING, "zstd_uncompress_dict:"
-                   " expects the second parameter to be string.");
-        RETURN_FALSE;
-    }
+    ZEND_PARSE_PARAMETERS_START(2, 2)
+        Z_PARAM_STRING(input, input_len)
+        Z_PARAM_STRING(dict, dict_len)
+    ZEND_PARSE_PARAMETERS_END();
 
-    unsigned long long const rSize = ZSTD_getDecompressedSize(Z_STRVAL_P(data),
-                                                              Z_STRLEN_P(data));
+    unsigned long long const rSize = ZSTD_getDecompressedSize(input,
+                                                              input_len);
     if (rSize == 0) {
-        zend_error(E_WARNING, "zstd_uncompress_dict:"
-                   " it was not compressed by zstd");
-        RETURN_FALSE;
-    }
-    void* const rBuff = emalloc((size_t)rSize);
-    if (!rBuff) {
-        zend_error(E_WARNING, "zstd_uncompress_dict: memory error");
+        ZSTD_WARNING("it was not compressed by zstd");
         RETURN_FALSE;
     }
 
     ZSTD_DCtx* const dctx = ZSTD_createDCtx();
     if (dctx == NULL) {
-        efree(rBuff);
-        zend_error(E_WARNING, "ZSTD_createDCtx() error");
+        ZSTD_WARNING("ZSTD_createDCtx() error");
         RETURN_FALSE;
     }
-    ZSTD_DDict* const ddict = ZSTD_createDDict(Z_STRVAL_P(dictBuffer),
-                                               Z_STRLEN_P(dictBuffer));
+    ZSTD_DDict* const ddict = ZSTD_createDDict(dict,
+                                               dict_len);
     if (!ddict) {
-        efree(rBuff);
-        zend_error(E_WARNING, "ZSTD_createDDict() error");
+        ZSTD_freeDStream(dctx);
+        ZSTD_WARNING("ZSTD_createDDict() error");
         RETURN_FALSE;
     }
-    size_t const dSize = ZSTD_decompress_usingDDict(dctx, rBuff, rSize,
-                                                    Z_STRVAL_P(data),
-                                                    Z_STRLEN_P(data),
+
+    output = zend_string_alloc(rSize, 0);
+
+    size_t const dSize = ZSTD_decompress_usingDDict(dctx, ZSTR_VAL(output), rSize,
+                                                    input,
+                                                    input_len,
                                                     ddict);
     if (dSize != rSize) {
-        efree(rBuff);
-        zend_error(E_WARNING, "zstd_uncompress_dict: %s",
-                   ZSTD_getErrorName(dSize));
+        ZSTD_freeDStream(dctx);
+        ZSTD_freeDDict(ddict);
+        zend_string_efree(output);
+        ZSTD_WARNING("%s", ZSTD_getErrorName(dSize));
         RETURN_FALSE;
     }
     ZSTD_freeDCtx(dctx);
     ZSTD_freeDDict(ddict);
 
-#if ZEND_MODULE_API_NO >= 20141001
-    RETVAL_STRINGL(rBuff, rSize);
-#else
-    RETVAL_STRINGL(rBuff, rSize, 1);
-#endif
-
-    efree(rBuff);
+    output = zstd_string_output_truncate(output, dSize);
+    RETVAL_NEW_STR(output);
 }
 
 
@@ -400,7 +395,7 @@ typedef struct _php_zstd_stream_data {
 
 #define STREAM_NAME "compress.zstd"
 
-static int php_zstd_decomp_close(php_stream *stream, int close_handle TSRMLS_DC)
+static int php_zstd_decomp_close(php_stream *stream, int close_handle)
 {
     STREAM_DATA_FROM_STREAM();
 
@@ -424,12 +419,12 @@ static int php_zstd_decomp_close(php_stream *stream, int close_handle TSRMLS_DC)
     return EOF;
 }
 
-static int php_zstd_comp_flush_or_end(php_zstd_stream_data *self, int end TSRMLS_DC)
+#if ZSTD_VERSION_NUMBER < 10400
+static int php_zstd_comp_flush_or_end(php_zstd_stream_data *self, int end)
 {
     size_t res;
     int ret = 0;
 
-#if ZSTD_VERSION_NUMBER < 10400
     /* Compress remaining data */
     if (self->input.size)  {
         self->input.pos = 0;
@@ -437,30 +432,26 @@ static int php_zstd_comp_flush_or_end(php_zstd_stream_data *self, int end TSRMLS
             self->output.size = self->sizeout;
             self->output.pos  = 0;
             res = ZSTD_compressStream(self->cctx, &self->output, &self->input);
-            if (ZSTD_isError(res)) {
-                php_error_docref(NULL TSRMLS_CC, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
+            if (ZSTD_IS_ERROR(res)) {
+                php_error_docref(NULL, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
                 ret = EOF;
             }
             php_stream_write(self->stream, self->bufout, self->output.pos);
         } while (self->input.pos != self->input.size);
     }
-#endif
 
     /* Flush / End */
     do {
         self->output.size = self->sizeout;
         self->output.pos  = 0;
-#if ZSTD_VERSION_NUMBER >= 10400
-        res = ZSTD_compressStream2(self->cctx, &self->output, &self->input, end ? ZSTD_e_end : ZSTD_e_flush);
-#else
+
         if (end) {
             res = ZSTD_endStream(self->cctx, &self->output);
         } else {
             res = ZSTD_flushStream(self->cctx, &self->output);
         }
-#endif
-        if (ZSTD_isError(res)) {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
+        if (ZSTD_IS_ERROR(res)) {
+            php_error_docref(NULL, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
             ret = EOF;
         }
         php_stream_write(self->stream, self->bufout, self->output.pos);
@@ -472,16 +463,39 @@ static int php_zstd_comp_flush_or_end(php_zstd_stream_data *self, int end TSRMLS
     return ret;
 }
 
+#else
+static int php_zstd_comp_flush_or_end(php_zstd_stream_data *self, int end)
+{
+    size_t res;
+    int ret = 0;
 
-static int php_zstd_comp_flush(php_stream *stream TSRMLS_DC)
+    ZSTD_inBuffer in = { NULL, 0, 0 };
+
+    /* Flush / End */
+    do {
+        self->output.pos  = 0;
+        res = ZSTD_compressStream2(self->cctx, &self->output, &in, end ? ZSTD_e_end : ZSTD_e_flush);
+        if (ZSTD_isError(res)) {
+            php_error_docref(NULL, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
+            ret = EOF;
+        }
+        php_stream_write(self->stream, self->output.dst, self->output.pos);
+    } while (res > 0);
+
+    return ret;
+}
+#endif
+
+
+static int php_zstd_comp_flush(php_stream *stream)
 {
     STREAM_DATA_FROM_STREAM();
 
-    return php_zstd_comp_flush_or_end(self, 0 TSRMLS_CC);
+    return php_zstd_comp_flush_or_end(self, 0);
 }
 
 
-static int php_zstd_comp_close(php_stream *stream, int close_handle TSRMLS_DC)
+static int php_zstd_comp_close(php_stream *stream, int close_handle)
 {
     STREAM_DATA_FROM_STREAM();
 
@@ -489,7 +503,7 @@ static int php_zstd_comp_close(php_stream *stream, int close_handle TSRMLS_DC)
         return EOF;
     }
 
-    php_zstd_comp_flush_or_end(self, 1 TSRMLS_CC);
+    php_zstd_comp_flush_or_end(self, 1);
 
     if (close_handle) {
         if (self->stream) {
@@ -499,8 +513,12 @@ static int php_zstd_comp_close(php_stream *stream, int close_handle TSRMLS_DC)
     }
 
     ZSTD_freeCCtx(self->cctx);
+#if ZSTD_VERSION_NUMBER >= 10400
+    efree(self->output.dst);
+#else
     efree(self->bufin);
     efree(self->bufout);
+#endif
     efree(self);
     stream->abstract = NULL;
 
@@ -509,11 +527,11 @@ static int php_zstd_comp_close(php_stream *stream, int close_handle TSRMLS_DC)
 
 
 #if PHP_VERSION_ID < 70400
-static size_t php_zstd_decomp_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+static size_t php_zstd_decomp_read(php_stream *stream, char *buf, size_t count)
 {
     size_t ret = 0;
 #else
-static ssize_t php_zstd_decomp_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+static ssize_t php_zstd_decomp_read(php_stream *stream, char *buf, size_t count)
 {
     ssize_t ret = 0;
 #endif
@@ -543,8 +561,8 @@ static ssize_t php_zstd_decomp_read(php_stream *stream, char *buf, size_t count
             self->output.pos = 0;
             self->output.size = self->sizeout;
             res = ZSTD_decompressStream(self->dctx, &self->output , &self->input);
-            if (ZSTD_isError(res)) {
-                php_error_docref(NULL TSRMLS_CC, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
+            if (ZSTD_IS_ERROR(res)) {
+                php_error_docref(NULL, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
 #if PHP_VERSION_ID >= 70400
                 return -1;
 #endif
@@ -567,18 +585,39 @@ static ssize_t php_zstd_decomp_read(php_stream *stream, char *buf, size_t count
 
 
 #if PHP_VERSION_ID < 70400
-static size_t php_zstd_comp_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
+static size_t php_zstd_comp_write(php_stream *stream, const char *buf, size_t count)
 {
     size_t ret = 0;
 #else
-static ssize_t php_zstd_comp_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
+static ssize_t php_zstd_comp_write(php_stream *stream, const char *buf, size_t count)
 {
     ssize_t ret = 0;
 #endif
-    size_t x, res;
 
     STREAM_DATA_FROM_STREAM();
 
+#if ZSTD_VERSION_NUMBER >= 10400
+    size_t res;
+    ZSTD_inBuffer in = { buf, count, 0 };
+
+    do {
+        self->output.pos = 0;
+        res = ZSTD_compressStream2(self->cctx, &self->output, &in, ZSTD_e_continue);
+        if (ZSTD_isError(res)) {
+            php_error_docref(NULL, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
+#if PHP_VERSION_ID >= 70400
+            return -1;
+#endif
+        }
+        php_stream_write(self->stream, self->output.dst, self->output.pos);
+
+    } while (res > 0);
+
+    return count;
+
+#else
+    size_t x, res;
+
     while(count > 0) {
         /* enough room for full data */
         if (self->input.size + count < self->sizein) {
@@ -602,13 +641,9 @@ static ssize_t php_zstd_comp_write(php_stream *stream, const char *buf, size_t c
         do {
             self->output.size = self->sizeout;
             self->output.pos  = 0;
-#if ZSTD_VERSION_NUMBER >= 10400
-            res = ZSTD_compressStream2(self->cctx, &self->output, &self->input, ZSTD_e_continue);
-#else
             res = ZSTD_compressStream(self->cctx, &self->output, &self->input);
-#endif
-            if (ZSTD_isError(res)) {
-                php_error_docref(NULL TSRMLS_CC, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
+            if (ZSTD_IS_ERROR(res)) {
+                php_error_docref(NULL, E_WARNING, "libzstd error %s\n", ZSTD_getErrorName(res));
 #if PHP_VERSION_ID >= 70400
                 return -1;
 #endif
@@ -620,6 +655,7 @@ static ssize_t php_zstd_comp_write(php_stream *stream, const char *buf, size_t c
         self->input.size = 0;
     }
     return ret;
+#endif
 }
 
 
@@ -652,21 +688,12 @@ static php_stream_ops php_stream_zstd_write_ops = {
 static php_stream *
 php_stream_zstd_opener(
     php_stream_wrapper *wrapper,
-#if PHP_VERSION_ID < 50600
-    char *path,
-    char *mode,
-#else
     const char *path,
     const char *mode,
-#endif
     int options,
-#if PHP_MAJOR_VERSION < 7
-    char **opened_path,
-#else
     zend_string **opened_path,
-#endif
     php_stream_context *context
-    STREAMS_DC TSRMLS_DC)
+    STREAMS_DC)
 {
     php_zstd_stream_data *self;
     int level = ZSTD_CLEVEL_DEFAULT;
@@ -683,7 +710,7 @@ php_stream_zstd_opener(
         }
     }
 
-    if (php_check_open_basedir(path TSRMLS_CC)) {
+    if (php_check_open_basedir(path)) {
         return NULL;
     }
 
@@ -692,12 +719,11 @@ php_stream_zstd_opener(
     } else if (!strcmp(mode, "r") || !strcmp(mode, "rb")) {
        compress = 0;
     } else {
-        php_error_docref(NULL TSRMLS_CC, E_ERROR, "zstd: invalid open mode");
+        php_error_docref(NULL, E_ERROR, "zstd: invalid open mode");
         return NULL;
     }
 
     if (context) {
-#if PHP_MAJOR_VERSION >= 7
         zval *tmpzval;
         zend_string *data;
 
@@ -714,29 +740,11 @@ php_stream_zstd_opener(
             }
             zend_string_release(data);
         }
-#endif
-#else
-        zval **tmpzval;
-
-        if (php_stream_context_get_option(context, "zstd", "level", &tmpzval) == SUCCESS) {
-            convert_to_long_ex(tmpzval);
-            level = Z_LVAL_PP(tmpzval);
-        }
-#if ZSTD_VERSION_NUMBER >= 10400
-        if (php_stream_context_get_option(context, "zstd", "dict", &tmpzval) == SUCCESS) {
-            convert_to_string(*tmpzval);
-            if (compress) {
-                cdict = ZSTD_createCDict(Z_STRVAL_PP(tmpzval), Z_STRLEN_PP(tmpzval), level);
-            } else {
-                ddict = ZSTD_createDDict(Z_STRVAL_PP(tmpzval), Z_STRLEN_PP(tmpzval));
-            }
-        }
-#endif
 #endif
     }
 
     if (level > ZSTD_maxCLevel()) {
-        php_error_docref(NULL TSRMLS_CC, E_WARNING, "zstd: compression level (%d) must be less than %d", level, ZSTD_maxCLevel());
+        php_error_docref(NULL, E_WARNING, "zstd: compression level (%d) must be less than %d", level, ZSTD_maxCLevel());
         level = ZSTD_maxCLevel();
     }
 
@@ -752,7 +760,7 @@ php_stream_zstd_opener(
         self->dctx = NULL;
         self->cctx = ZSTD_createCCtx();
         if (!self->cctx) {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING, "zstd: compression context failed");
+            php_error_docref(NULL, E_WARNING, "zstd: compression context failed");
             php_stream_close(self->stream);
             efree(self);
             return NULL;
@@ -761,9 +769,14 @@ php_stream_zstd_opener(
         ZSTD_CCtx_reset(self->cctx, ZSTD_reset_session_only);
         ZSTD_CCtx_refCDict(self->cctx, cdict);
         ZSTD_CCtx_setParameter(self->cctx, ZSTD_c_compressionLevel, level);
+
+        self->output.size = ZSTD_CStreamOutSize();
+        self->output.dst  = emalloc(self->output.size);
+        self->output.pos  = 0;
+
 #else
         ZSTD_initCStream(self->cctx, level);
-#endif
+
         self->bufin = emalloc(self->sizein = ZSTD_CStreamInSize());
         self->bufout = emalloc(self->sizeout = ZSTD_CStreamOutSize());
         self->input.src  = self->bufin;
@@ -772,13 +785,14 @@ php_stream_zstd_opener(
         self->output.dst = self->bufout;
         self->output.pos  = 0;
         self->output.size = 0;
+#endif
 
         return php_stream_alloc(&php_stream_zstd_write_ops, self, NULL, mode);
 
     } else {
         self->dctx = ZSTD_createDCtx();
         if (!self->dctx) {
-            php_error_docref(NULL TSRMLS_CC, E_WARNING, "zstd: compression context failed");
+            php_error_docref(NULL, E_WARNING, "zstd: compression context failed");
             php_stream_close(self->stream);
             efree(self);
             return NULL;
@@ -801,7 +815,6 @@ php_stream_zstd_opener(
 
         return php_stream_alloc(&php_stream_zstd_read_ops, self, NULL, mode);
     }
-    return NULL;
 }
 
 
@@ -815,10 +828,8 @@ static php_stream_wrapper_ops zstd_stream_wops = {
     NULL,    /* unlink */
     NULL,    /* rename */
     NULL,    /* mkdir */
-    NULL    /* rmdir */
-#if PHP_VERSION_ID >= 50400
-    , NULL
-#endif
+    NULL,    /* rmdir */
+    NULL
 };
 
 
@@ -828,7 +839,7 @@ php_stream_wrapper php_stream_zstd_wrapper = {
     0 /* is_url */
 };
 
-#if PHP_MAJOR_VERSION >= 7 && defined(HAVE_APCU_SUPPORT)
+#if defined(HAVE_APCU_SUPPORT)
 static int APC_SERIALIZER_NAME(zstd)(APC_SERIALIZER_ARGS)
 {
     int result;
@@ -845,10 +856,6 @@ static int APC_SERIALIZER_NAME(zstd)(APC_SERIALIZER_ARGS)
 
     size = ZSTD_compressBound(ZSTR_LEN(var.s));
     *buf = (char*) emalloc(size + 1);
-    if (*buf == NULL) {
-        *buf_len = 0;
-        return 0;
-    }
 
     *buf_len = ZSTD_compress(*buf, size, ZSTR_VAL(var.s), ZSTR_LEN(var.s),
                              DEFAULT_COMPRESS_LEVEL);
@@ -883,10 +890,6 @@ static int APC_UNSERIALIZER_NAME(zstd)(APC_UNSERIALIZER_ARGS)
     }
 
     var = emalloc(size);
-    if (!var) {
-        ZVAL_NULL(value);
-        return 0;
-    }
 
     var_len = ZSTD_decompress(var, size, buf, buf_len);
     if (ZSTD_isError(var_len) || var_len == 0) {
@@ -936,9 +939,9 @@ ZEND_MINIT_FUNCTION(zstd)
                            ZSTD_VERSION_STRING,
                            CONST_CS | CONST_PERSISTENT);
 
-    php_register_url_stream_wrapper(STREAM_NAME, &php_stream_zstd_wrapper TSRMLS_CC);
+    php_register_url_stream_wrapper(STREAM_NAME, &php_stream_zstd_wrapper);
 
-#if PHP_MAJOR_VERSION >= 7 && defined(HAVE_APCU_SUPPORT)
+#if defined(HAVE_APCU_SUPPORT)
     apc_register_serializer("zstd",
                             APC_SERIALIZER_NAME(zstd),
                             APC_UNSERIALIZER_NAME(zstd),
@@ -954,7 +957,7 @@ ZEND_MINFO_FUNCTION(zstd)
     php_info_print_table_row(2, "Zstd support", "enabled");
     php_info_print_table_row(2, "Extension Version", PHP_ZSTD_EXT_VERSION);
     php_info_print_table_row(2, "Interface Version", ZSTD_VERSION_STRING);
-#if PHP_MAJOR_VERSION >= 7 && defined(HAVE_APCU_SUPPORT)
+#if defined(HAVE_APCU_SUPPORT)
     php_info_print_table_row(2, "APCu serializer ABI", APC_SERIALIZER_ABI);
 #endif
     php_info_print_table_end();
@@ -976,8 +979,6 @@ static zend_function_entry zstd_functions[] = {
     ZEND_FALIAS(zstd_decompress_usingcdict,
                 zstd_uncompress_dict, arginfo_zstd_uncompress_dict)
 
-// PHP 5.3+
-#if ZEND_MODULE_API_NO >= 20090626
     ZEND_NS_FALIAS(PHP_ZSTD_NS, compress,
                    zstd_compress, arginfo_zstd_compress)
     ZEND_NS_FALIAS(PHP_ZSTD_NS, uncompress,
@@ -996,11 +997,11 @@ static zend_function_entry zstd_functions[] = {
                    zstd_uncompress_dict, arginfo_zstd_uncompress_dict)
     ZEND_NS_FALIAS(PHP_ZSTD_NS, decompress_usingcdict,
                    zstd_uncompress_dict, arginfo_zstd_uncompress_dict)
-#endif
+
     {NULL, NULL, NULL}
 };
 
-#if PHP_MAJOR_VERSION >= 7 && defined(HAVE_APCU_SUPPORT)
+#if defined(HAVE_APCU_SUPPORT)
 static const zend_module_dep zstd_module_deps[] = {
     ZEND_MOD_OPTIONAL("apcu")
     ZEND_MOD_END
@@ -1008,11 +1009,11 @@ static const zend_module_dep zstd_module_deps[] = {
 #endif
 
 zend_module_entry zstd_module_entry = {
-#if PHP_MAJOR_VERSION >= 7 && defined(HAVE_APCU_SUPPORT)
+#if defined(HAVE_APCU_SUPPORT)
     STANDARD_MODULE_HEADER_EX,
     NULL,
     zstd_module_deps,
-#elif ZEND_MODULE_API_NO >= 20010901
+#else
     STANDARD_MODULE_HEADER,
 #endif
     "zstd",
@@ -1022,9 +1023,7 @@ zend_module_entry zstd_module_entry = {
     NULL,
     NULL,
     ZEND_MINFO(zstd),
-#if ZEND_MODULE_API_NO >= 20010901
     PHP_ZSTD_EXT_VERSION,
-#endif
     STANDARD_MODULE_PROPERTIES
 };
 
